----------------------------------------------------------------------
----                                                              ----
---- Atari STE compatible IP Core				                  ----
----                                                              ----
---- This file is part of the SUSKA ATARI clone project.          ----
---- http://www.experiment-s.de                                   ----
----                                                              ----
---- Description:                                                 ----
---- This a small test configuration for the flash boot loader.   ----
----                                                              ----
----------------------------------------------------------------------
----                                                              ----
---- Copyright (C) 2007 Wolfgang Foerster                         ----
----                                                              ----
---- This source file may be used and distributed without         ----
---- restriction provided that this copyright statement is not    ----
---- removed from the file and that any derivative work contains  ----
---- the original copyright notice and the associated disclaimer. ----
----                                                              ----
---- This source file is free software; you can redistribute it   ----
---- and/or modify it under the terms of the GNU Lesser General   ----
---- Public License as published by the Free Software Foundation; ----
---- either version 2.1 of the License, or (at your option) any   ----
---- later version.                                               ----
----                                                              ----
---- This source is distributed in the hope that it will be       ----
---- useful, but WITHOUT ANY WARRANTY; without even the implied   ----
---- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ----
---- PURPOSE. See the GNU Lesser General Public License for more  ----
---- details.                                                     ----
----                                                              ----
---- You should have received a copy of the GNU Lesser General    ----
---- Public License along with this source; if not, download it   ----
---- from http://www.gnu.org/licenses/lgpl.html                   ----
----                                                              ----
----------------------------------------------------------------------
-- 
-- Revision History
-- 
-- Revision 2K6B  2007/12/20 WF
--   Initial Release.

library work;
use work.wf_suska_pkg.all;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity WF_SUSKA_TOP is
	port(
		-- System controls:
		RESET_COREn			: in bit; -- FPGA reset.
		RESETn				: in std_logic; -- System reset.
		CLK_PLL1			: in std_logic; -- 16 MHz system clock.

		-- Data and address busses:
		DATA				: inout std_logic_vector(15 downto 0);
		ADR					: inout std_logic_vector(23 downto 1);

		-- Flash controls:
		FLASH_RDY			: in bit;
		FLASH_ADR_19		: out bit;
		FLASH_ADR_18		: out bit;
		FLASH_RESETn		: out bit;
		FLASH_WEn			: out bit;
		FLASH_OEn			: out bit;
		FLASH_CEn			: out bit;

		-- Configuration switches:
		-- Use weak the pull up feature or pull ups. Connect the switches
		-- to GND.
		FLASH_SEL			: in bit_vector(1 downto 0); -- ROM space selection.

		-- System and status:
		SDC_AVR_CLK			: out std_logic; -- Boot microcontroller clock.
		PLL_FAULT			: out bit; -- Indicates unlocked PLLs.
		BOOT_LED			: out bit; -- Boot loader active...
		
		TEST1			: out bit; -- ROM select connector X19 pin 1.
		TEST2			: out bit; -- ROM select connector X19 pin 3.
		TEST3			: out bit; -- ROM select connector X19 pin 5.
		
		-- Microcontroller interface:
		MC_SPI_CLK			: in bit;
		MC_SPI_DIN			: buffer bit;
		MC_SPI_DOUT			: in bit;
		BOOT_ACK			: in bit;
		BOOT_REQ			: out bit
	);
end entity WF_SUSKA_TOP;

architecture STRUCTURE of WF_SUSKA_TOP is
-- Hardware specific components:
component cyclone2_vram
	PORT
	(
		address		: IN STD_LOGIC_VECTOR (14 DOWNTO 0);
		clock		: IN STD_LOGIC ;
		data		: IN STD_LOGIC_VECTOR (7 DOWNTO 0);
		wren		: IN STD_LOGIC ;
		q			: OUT STD_LOGIC_VECTOR (7 DOWNTO 0)
	);
end component;

component cyclone2_pll_1
	PORT
	(
		areset		: IN STD_LOGIC  := '0';
		clkswitch	: IN STD_LOGIC  := '0';
		inclk0		: IN STD_LOGIC  := '0';
		inclk1		: IN STD_LOGIC  := '0';
		c0			: OUT STD_LOGIC ;
		c1			: OUT STD_LOGIC ;
		c2			: OUT STD_LOGIC ;
		locked		: OUT STD_LOGIC 
	);
end component;

component cyclone2_pll_2
	PORT
	(
		areset		: IN STD_LOGIC  := '0';
		inclk0		: IN STD_LOGIC  := '0';
		c0			: OUT STD_LOGIC ;
		c1			: OUT STD_LOGIC ;
		c2			: OUT STD_LOGIC ;
		locked		: OUT STD_LOGIC 
	);
end component;
signal RESET_Sn			: bit;
signal RESET_CORE_Sn	: bit;
signal IDE_RES_In		: bit;
signal PLL1_LOCKED		: std_logic;
signal CLK_PLL_4x		: std_logic;
signal CLK_PLL_2x		: std_logic;
signal CLK_PLL_1x		: std_logic;
signal CLK_4			: bit;
signal CLK_2			: bit;
signal CLK_1			: bit;
signal DATA_OUT_BOOT	: std_logic_vector(15 downto 0);
signal DATA_EN_BOOT		: bit;
signal ADR_OUT_BOOT		: std_logic_vector(20 downto 1);
signal ADR_EN_BOOT		: bit;
signal PLL_LOCKS		: bit;
begin
	TEST1 <= MC_SPI_CLK;
	TEST2 <= MC_SPI_DIN;
	TEST3 <= MC_SPI_DOUT;

	KEY_SCAN: process
	-- Sample the RESETn and the RESET_COREn buttons
	-- about every 5ms. This provides stability against
	-- push button jitter.
	variable SCAN_TIMER	: std_logic_vector(16 downto 0);
	begin
		wait until CLK_1 = '1' and CLK_1' event;
		if SCAN_TIMER <= '1' & x"3880" then
			SCAN_TIMER := SCAN_TIMER + '1';
		else
			SCAN_TIMER := (others => '0');
			RESET_Sn <= To_Bit(RESETn);
			RESET_CORE_Sn <= RESET_COREn;
		end if;
	end process KEY_SCAN;

	PLL_FAULT <= '1' when PLL1_LOCKED = '0' else '0'; 
	PLL_LOCKS <= '1' when PLL1_LOCKED = '1' else '0';

	I_SYSCLOCKS: cyclone2_pll_1
		port map(
			areset				=> '0',
			clkswitch			=> '0',
			inclk0				=> CLK_PLL1, -- 16MHz.
			inclk1				=> '0',
			c0					=> CLK_PLL_1x, -- 16MHz.
			c1					=> CLK_PLL_2x, -- 32MHz.
			c2					=> CLK_PLL_4x, -- 64MHz.
			locked				=> PLL1_LOCKED
		);

	CLK_4 <= To_Bit(CLK_PLL_4x);
	CLK_2 <= To_Bit(CLK_PLL_2x);
	CLK_1 <= To_Bit(CLK_PLL_1x);
	SDC_AVR_CLK <= CLK_PLL_1x; -- 16MHz.

	ADR <= "000" & ADR_OUT_BOOT when ADR_EN_BOOT = '1' else (others => 'Z');
	DATA <= DATA_OUT_BOOT when DATA_EN_BOOT = '1' else (others => 'Z');

	-- Flash adress space select: 
	-- The original TOS operating system takes only the space of 18
	-- address lines. Thus, the flash device is capable of storing
	-- four operating system versions selectable via FLAHS_SEL.
	FLASH_ADR_19 <= To_Bit(ADR_OUT_BOOT(19)) when ADR_EN_BOOT = '1' else not FLASH_SEL(1); -- Inverted due to swicthes to GND.
	FLASH_ADR_18 <= To_Bit(ADR_OUT_BOOT(18)) when ADR_EN_BOOT = '1' else not FLASH_SEL(0); -- Inverted due to swicthes to GND.
	
	I_FLASHBOOT: WF_FLASHBOOT
		port map(
			CLK						=> CLK_1, -- 16MHz.
			PLL_LOCK				=> PLL_LOCKS,
			RESET_COREn				=> RESET_CORE_Sn,
			RESET_INn				=> RESET_Sn,
			-- RESET_OUTn			=> Not used.
			ROM_CEn					=> '0',
			ADR_OUT(19 downto 0)	=> ADR_OUT_BOOT, -- High address bits currently not in use.
			ADR_EN					=> ADR_EN_BOOT,
			DATA_IN					=> DATA,
			DATA_OUT				=> DATA_OUT_BOOT,
			DATA_EN					=> DATA_EN_BOOT,
			FLASH_RDY				=> FLASH_RDY,
			FLASH_RESETn			=> FLASH_RESETn,
			FLASH_WEn				=> FLASH_WEn,
			FLASH_OEn				=> FLASH_OEn,
			FLASH_CEn				=> FLASH_CEn,
			SPI_CLK					=> MC_SPI_CLK,
			SPI_DIN					=> MC_SPI_DOUT,
			SPI_DOUT				=> MC_SPI_DIN,
			BOOT_ACK				=> BOOT_ACK,
			BOOT_REQ				=> BOOT_REQ,
			BOOT_LED				=> BOOT_LED
		);
end architecture STRUCTURE;
